#define UNICODE 1
#define WIN32_LEAN_AND_MEAN 1

#include <windows.h>

#include <wchar.h>
#include <stdlib.h>
#include <stdio.h>
#include <ctype.h>
#include <limits.h>
#include <errno.h>

void ReportError(wchar_t* str, DWORD dwError)
{
	if (str != NULL && *str != L'\0')
		fwprintf(stderr, L"%s: ", str);
	wchar_t* message;
	if (!FormatMessage(FORMAT_MESSAGE_ALLOCATE_BUFFER | FORMAT_MESSAGE_FROM_SYSTEM | FORMAT_MESSAGE_IGNORE_INSERTS,
		0, dwError, 0, (wchar_t*)&message, 0, NULL))
	{
		fwprintf(stderr, L"Error %d\n", (int)dwError);
		return;
	}
	fwprintf(stderr, L"%s\n", message);
	LocalFree(message);
}

int wmain(int argc, wchar_t** argv)
{
	DWORD dwError;

	int nonoption_count = 0;
	wchar_t* nonoption[2];
	bool allow_truncate = false;
	bool usage = false;
	for (++argv; *argv != NULL; ++argv)
	{
		if (**argv == L'/')
		{
			if (_wcsicmp(*argv, L"/allowtruncate") == 0)
				allow_truncate = true;
			else
			{
				usage = true;
				break;
			}
		}
		else
		{
			if (nonoption_count >= 2)
			{
				usage = true;
				break;
			}
			nonoption[nonoption_count++] = *argv;
		}
	}
	wchar_t* end;
	__int64 size;
	if (usage || nonoption_count != 2 || !isdigit(*nonoption[1]) || (size = _wcstoi64(nonoption[1], &end, 0), *end != L'\0'))
	{
		fputws(L"Usage: CreateZeroFile.exe [/allowtruncate] filename size\n", stderr);
		return 1;
	}
	if (size == _I64_MAX && errno == ERANGE)
	{
		fputws(L"Size is too large\n", stderr);
		return 1;
	}
	wchar_t* filename = nonoption[0];
	HANDLE hFile = CreateFile(filename, GENERIC_WRITE, 0, NULL, OPEN_ALWAYS, FILE_ATTRIBUTE_NORMAL, NULL);
	if (hFile == INVALID_HANDLE_VALUE)
	{
		dwError = GetLastError();
		ReportError(filename, dwError);
		return 1;
	}
	LARGE_INTEGER current_size;
	if ((current_size.LowPart = GetFileSize(hFile, (LPDWORD)&current_size.HighPart)) == INVALID_FILE_SIZE &&
		(dwError = GetLastError()) != NO_ERROR)
	{
		ReportError(filename, dwError);
		return 1;
	}
	if (!allow_truncate && size < current_size.QuadPart)
	{
		fputws(L"Truncation is not allowed; specify /allowtruncate to force\n", stderr);
		return 1;
	}
	LARGE_INTEGER size_li;
	size_li.QuadPart = size;
	if (SetFilePointer(hFile, size_li.LowPart, &size_li.HighPart, FILE_BEGIN) == INVALID_SET_FILE_POINTER &&
		(dwError = GetLastError()) != NO_ERROR)
	{
		ReportError(filename, dwError);
		return 1;
	}
	if (!SetEndOfFile(hFile))
	{
		dwError = GetLastError();
		ReportError(filename, dwError);
		return 1;
	}
	if (!CloseHandle(hFile))
	{
		dwError = GetLastError();
		ReportError(filename, dwError);
		return 1;
	}
	return 0;
}
